
import * as amqp from 'amqplib';
import * as jobTracker from './job-tracker';
import {
  promiseWithTimeout,
  TimeoutError,
} from './util';
const uuid = require('uuid/v4');
const debug = require('debug')('minion-kit:queue')

const CONNECTION_COUNT = 1
const CHANNEL_COUNT_PER_CONNECTION = 3
const PENDING_JOB_TTL = 60


export class Queue {

  connectionString
  connectionCount
  channelCountPerConnection
  pendingJobTTL
  currentChannelIdx = 0
  connections = []
  consumerChannels = []
  channels = []


  constructor (setup = {}) {
    const {
      connectionString,
      connectionCount = CONNECTION_COUNT,
      channelCountPerConnection = CHANNEL_COUNT_PER_CONNECTION,
      pendingJobTTL = PENDING_JOB_TTL,
    } = setup

    this.connectionString = connectionString
    this.connectionCount = connectionCount
    this.channelCountPerConnection = channelCountPerConnection
    this.pendingJobTTL = pendingJobTTL
  }


  connect = async () => {

    const conList = Array( this.connectionCount ).fill(0)
    const chList = Array( this.channelCountPerConnection ).fill(0)

    await Promise.all(conList.map(async (con, conIdx) => {
      const newConnection = await amqp.connect( this.connectionString )
      this.connections.push(newConnection)
      debug(`Established connection #${conIdx} with message queue`)

      await Promise.all(chList.map(async (ch, chIdx) => {
        const newChannel = await newConnection.createConfirmChannel()
        this.channels.push( newChannel )
        debug(`Opened channel #${chIdx} on connection #${conIdx} with message queue (total: #${this.channels.length})`)
      }))
    }))
  }



  disconnect = async () => {

    await Promise.all(this.channels.map(async (ch, chIdx) => {
      debug(`Closing channel #${chIdx}`)
      await ch.waitForConfirms()
    }))

    await Promise.all(this.consumerChannels.map(async (ch, chIdx) => {
      debug(`Closing consumer channel #${chIdx}`)
      await ch.waitForConfirms()
    }))

    await Promise.all(this.connections.map(async (con, conIdx) => {
      debug(`Closing connection #${conIdx}`)
      await con.close()
    }))
  }



  getChannel = () => {

    this.currentChannelIdx++

    if (this.currentChannelIdx >= this.channels.length) {
      this.currentChannelIdx = 0
    }

    return this.channels[ this.currentChannelIdx ]
  }



  sendToQueue = async (queueName, payload, options) => {

    const channel = this.getChannel()
    await channel.assertQueue(queueName, { durable: true, noAck: false })

    debug(`Sending new job to queue ${queueName}`)

    return channel.sendToQueue(
      queueName,
      Buffer.from(JSON.stringify(payload)),
      {
        persistent: true,
        contentType: 'application/json',
        headers: options
      },
    )

  }



  requeueJob = async (consumerChannel, queueName, payload, msg) => {

    if ( (msg.properties.headers.attemptsCount + 1) < msg.properties.headers.attemptsLimit) {

      await this.sendToQueue(queueName, payload, {
        attemptsLimit: msg.properties.headers.attemptsLimit,
        attemptsCount: msg.properties.headers.attemptsCount + 1,
        finalQueueOnAttemptsExceeded: msg.properties.headers.finalQueueOnAttemptsExceeded
      })

      consumerChannel.nack(msg, false, false);
      debug(`[${payload.correlationId}] Requeued job`);
    }
    else {
      if (msg.properties.headers.finalQueueOnAttemptsExceeded) {
        const finalQueueName = msg.properties.headers.finalQueueOnAttemptsExceeded

        await this.sendToQueue(finalQueueName, payload, {
          attemptsCount: msg.properties.headers.attemptsCount,
        })

        consumerChannel.nack(msg, false, false);
        debug(`[${payload.correlationId}] Sent job to final queue`);
      }
      else {
        consumerChannel.nack(msg, false, false);
      }
    }
  }



  consume = async (consumerQueueName, jobProcessor) => {

    const newConnection = await amqp.connect( this.connectionString )
    debug('Established consumer connection with message queue')

    const consumerChannel = await newConnection.createConfirmChannel()
    consumerChannel.prefetch(1);
    this.consumerChannels.push(consumerChannel)
    debug(`Established consumer channel with queue ${consumerQueueName}`)

    await consumerChannel.assertQueue(consumerQueueName, { durable: true, noAck: false })

    consumerChannel.consume(consumerQueueName, (msg) => {

      let payload

      try {
        payload = JSON.parse(msg.content)

        if (typeof payload !== 'object') {
          throw new Error('Message should be an object')
        }

      }
      catch(err) {
        debug({err: err}, 'Invalid message format')
        return consumerChannel.nack(msg, false, false);
      }

      if (!payload.correlationId) {
        payload.correlationId = uuid()
      }

      debug(`[${payload.correlationId}] Fetched job`);

      jobTracker.trackPendingJobs( [ payload ] )

      // set defaults
      msg.properties.headers = msg.properties.headers || {}
      msg.properties.headers.attemptsLimit = msg.properties.headers.attemptsLimit || 0
      msg.properties.headers.attemptsCount = msg.properties.headers.attemptsCount || 0

      const jobProcessorPromise = new Promise((resolve, reject) => {
        try {
          resolve(jobProcessor(payload, consumerQueueName))
        }
        catch (err) {
          reject(err)
        }
      })

      return promiseWithTimeout(this.pendingJobTTL * 1000, jobProcessorPromise)

        .then(() => {
          jobTracker.increaseStatusCompleted()
          consumerChannel.ack(msg);
          debug(`[${payload.correlationId}] Completed job`);
        })

        .catch((err) => {
          if (err instanceof TimeoutError) {
            jobTracker.increaseStatusTimedOut()
          }
          else {
            jobTracker.increaseStatusFailed()
          }

          debug({err: err, correlationId: payload.correlationId})
          this.requeueJob(consumerChannel, consumerQueueName, payload, msg)
        })

        .then(() => {
          jobTracker.unTrackPendingJob(payload)
        })

    })
  }

}
